"""Unit tests for module fixed_shift_dask"""

import numpy as np
import pytest

from lisainstrument.dynamic_delay_dask import (
    make_dynamic_shift_lagrange_dask,
    numpyfy_dask_multi,
)
from lisainstrument.dynamic_delay_numpy import ShiftBC
from lisainstrument.fixed_shift_dask import make_fixed_shift_lagrange_dask
from lisainstrument.fixed_shift_numpy import make_fixed_shift_lagrange_numpy


def g(x):
    """This example functionis a polynomial of order 5, which should be reproduced
    exactly when being interpolated with Lagrange polynomials of order 5.
    """
    return (
        4.32546
        + 3.34324 * x
        - 4.342 * x**2
        - 0.46 * x**3
        + 1.43598 * x**4
        - np.pi * x**5
    )


def test_fixed_shift_lagrange_dask() -> None:
    """Check that dask-based Lagrange interpolator for fixed shifts works.

    Compares dask-based interpolator against numpy based version and against
    expected result (excluding points affected by boundary treatment).
    """
    order = 5

    t, dt = np.linspace(-5.345, 10.345, 1003, retstep=True, dtype=float)
    y = g(t)

    for d in (0.93456456 / dt, 3.1, 3.5, 3.9, 4.0, 4.1, 4.5, 4.9, 116.0):

        op_np = make_fixed_shift_lagrange_numpy(ShiftBC.FLAT, ShiftBC.EXCEPTION, order)
        op_da = make_fixed_shift_lagrange_dask(ShiftBC.FLAT, ShiftBC.EXCEPTION, order)
        op_na = numpyfy_dask_multi(op_da, chunks=113)

        s_np = op_np(y, -d)
        s_na = op_na(y, -d)
        s_ex = g(t - d * dt)

        margin_ex = -int(np.floor(-d)) + order // 2

        assert s_na == pytest.approx(s_np, rel=1e-14, abs=0)
        assert s_ex[margin_ex:] == pytest.approx(s_na[margin_ex:], abs=1e-15, rel=5e-13)

        with pytest.raises(RuntimeError):
            op_na(y, d)

        op2_np = make_fixed_shift_lagrange_numpy(ShiftBC.EXCEPTION, ShiftBC.FLAT, order)
        op2_da = make_fixed_shift_lagrange_dask(ShiftBC.EXCEPTION, ShiftBC.FLAT, order)
        op2_na = numpyfy_dask_multi(op2_da, chunks=113)

        s2_np = op2_np(y, d)
        s2_na = op2_na(y, d)
        s2_ex = g(t + d * dt)

        margin2_ex = int(np.floor(d)) + (order - order // 2)

        assert s2_na == pytest.approx(s2_np, rel=1e-14, abs=0)
        assert s2_ex[:-margin2_ex] == pytest.approx(
            s2_na[:-margin2_ex], abs=1e-15, rel=5e-13
        )

        with pytest.raises(RuntimeError):
            op2_na(y, -d)

        op3_da = make_fixed_shift_lagrange_dask(ShiftBC.FLAT, ShiftBC.FLAT, order)
        op3_na = numpyfy_dask_multi(op3_da, chunks=113)

        s3_na = op3_na(y, -d)
        assert s3_na == pytest.approx(s_np, rel=1e-14, abs=0)

        s4_na = op3_na(y, d)
        assert s4_na == pytest.approx(s2_np, rel=1e-14, abs=0)


def test_shift_lagrange_dask_fixed_vs_dyn() -> None:
    """Check that dask-based fixed and dynamic lagrange shift functions yields identical results

    Uses different interpolation orders (odd, even, high) and shifts. Shift values chosen
    to take into account use of round and floor functions in the code to be tested.
    """
    t, dt = np.linspace(-5.345, 10.345, 1003, retstep=True, dtype=float)
    y = g(t)

    for order in (5, 6, 31):

        for d in (0.93456456 / dt, 23.1, 23.5, 23.9, 24.0, 116.0):
            d1d = np.ones_like(y) * d

            op1_da = make_dynamic_shift_lagrange_dask(
                order, d, d, ShiftBC.FLAT, ShiftBC.EXCEPTION
            )
            op1_na = numpyfy_dask_multi(op1_da, chunks=113)

            op2_da = make_fixed_shift_lagrange_dask(
                ShiftBC.FLAT, ShiftBC.EXCEPTION, order
            )
            op2_na = numpyfy_dask_multi(op2_da, chunks=113)

            s1_np = op1_na(y, -d1d)
            s2_np = op2_na(y, -d)

            assert s1_np == pytest.approx(s2_np, rel=1e-14, abs=0)
